@@ -1,8 +1,7 @@
-## (c) Simon N. Wood (2013,2014). Provided under GPL 2.
+## (c) Simon N. Wood (2013-2015). Provided under GPL 2.
 ## Routines for gam estimation beyond exponential family.
 
 
-
 dDeta <- function(y,mu,wt,theta,fam,deriv=0) {
 ## What is available directly from the family are derivatives of the 
 ## deviance and link w.r.t. mu. This routine converts these to the
@@ -203,6 +202,11 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
 ## sp contains the vector of extended family parameters, followed by the log smoothing parameters,
 ## followed by the log scale parameter if scale < 0
 
+  ## some families have second derivative of deviance, and hence iterative weights
+  ## very close to zero for some data. This can lead to poorly scaled sqrt(w)z
+  ## and it is better to base everything on wz...
+  if (is.null(family$use.wz)) family$use.wz <- FALSE
+
   if (family$n.theta>0) { ## there are extra parameters to estimate
     ind <- 1:family$n.theta
     theta <- sp[ind] ## parameters of the family
@@ -266,6 +270,21 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
   if (is.null(weights)) weights <- rep.int(1, nobs)
   if (is.null(offset)) offset <- rep.int(0, nobs)
 
+  linkinv <- family$linkinv
+  valideta <- family$valideta
+  validmu <- family$validmu
+  dev.resids <- family$dev.resids
+
+  ## need an initial `null deviance' to test for initial divergence...
+  ## if (!is.null(start)) null.coef <- start - can be on edge of feasible - not good
+  null.eta <- as.numeric(x%*%null.coef + as.numeric(offset))
+  old.pdev <- sum(dev.resids(y, linkinv(null.eta), weights,theta)) + t(null.coef)%*%St%*%null.coef 
+
+  if (!is.null(start)) { ## check it's at least better than null.coef
+    pdev <- sum(dev.resids(y, linkinv(x%*%start+as.numeric(offset)), weights,theta)) + t(start)%*%St%*%start
+    if (pdev>old.pdev) start <- mustart <- etastart <- NULL
+  }
+
   ## call the families initialization code...
 
   if (is.null(mustart)) {
@@ -295,42 +314,49 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
    ##mu.eta <- family$mu.eta
    ##Dd <- family$Dd
 
-   linkinv <- family$linkinv
-   valideta <- family$valideta
-   validmu <- family$validmu
-   dev.resids <- family$dev.resids
  
    mu <- linkinv(eta);etaold <- eta
      
    ## need an initial `null deviance' to test for initial divergence...
    ## if (!is.null(start)) null.coef <- start - can be on edge of feasible - not good
+   #null.eta <- as.numeric(x%*%null.coef + as.numeric(offset))
+   #old.pdev <- sum(dev.resids(y, linkinv(null.eta), weights,theta)) + t(null.coef)%*%St%*%null.coef 
+
    coefold <- null.coef
-   null.eta <- as.numeric(x%*%null.coef + as.numeric(offset))
-   old.pdev <- sum(dev.resids(y, linkinv(null.eta), weights,theta)) + t(null.coef)%*%St%*%null.coef 
    conv <-  boundary <- FALSE
  
    for (iter in 1:control$maxit) { ## start of main fitting iteration 
       if (control$trace) cat(iter," ")
       dd <- dDeta(y,mu,weights,theta,family,0) ## derivatives of deviance w.r.t. eta
-     
-      good <- is.finite(dd$Deta.Deta2)
-      if (control$trace&sum(!good)>0) cat("\n",sum(!good)," not good\n") 
+
+      # good <- is.finite(dd$Deta.Deta2)
+  
       w <- dd$Deta2 * .5;
+      wz <- w*(eta-offset) - .5*dd$Deta
+      z <- (eta-offset) - dd$Deta.Deta2
+      good <- is.finite(z)&is.finite(w)
+      if (control$trace&sum(!good)>0) cat("\n",sum(!good)," not good\n")
       if (sum(!good)) {
-        good1 <- is.finite(w)&good ## make sure w finite too
-        w[!is.finite(w)] <- 0      ## clear infinite w
-        w[!good1&w==0] <- max(w)*.Machine$double.eps^.5 ## reset zero value weights for problem elements
-        dd$Deta.Deta2[!good] <- .5*dd$Deta[!good]/w[!good] ## reset problem elements to finite
-        good <- is.finite(dd$Deta.Deta2) ## check in case Deta not finite, for example
-      }
-      z <- (eta-offset)[good] - dd$Deta.Deta2[good] ## - .5 * dd$Deta[good] / w
+        use.wy <- TRUE
+        good <- is.finite(w)&is.finite(wz)
+        z[!is.finite(z)] <- 0 ## avoid NaN in .C call - unused anyway
+      } else use.wy <- family$use.wz
       
-      oo <- .C(C_pls_fit1,   ##C_pls_fit1, reinstate for use in mgcv
-               y=as.double(z),X=as.double(x[good,]),w=as.double(w),
+      #if (sum(!good)) {
+      #  good1 <- is.finite(w)&good ## make sure w finite too
+      #  w[!is.finite(w)] <- 0      ## clear infinite w
+      #  w[!good1&w==0] <- max(w)*.Machine$double.eps^.5 ## reset zero value weights for problem elements
+      #  dd$Deta.Deta2[!good] <- .5*dd$Deta[!good]/w[!good] ## reset problem elements to finite
+      #  good <- is.finite(dd$Deta.Deta2) ## check in case Deta not finite, for example
+      #}
+      #z <- (eta-offset)[good] - dd$Deta.Deta2[good] ## - .5 * dd$Deta[good] / w
+      
+      oo <- .C(C_pls_fit1,   
+               y=as.double(z[good]),X=as.double(x[good,]),w=as.double(w[good]),wy = as.double(wz[good]),
                      E=as.double(Sr),Es=as.double(Eb),n=as.integer(sum(good)),
                      q=as.integer(ncol(x)),rE=as.integer(rows.E),eta=as.double(z),
                      penalty=as.double(1),rank.tol=as.double(rank.tol),
-                     nt=as.integer(control$nthreads))
+                     nt=as.integer(control$nthreads),use.wy=as.integer(use.wy))
       if (oo$n<0) { ## then problem is indefinite - switch to +ve weights for this step
         if (control$trace) cat("**using positive weights\n")
         # problem is that Fisher can be very poor for zeroes  
@@ -338,18 +364,27 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
         ## index weights that are finite and positive 
         good <- is.finite(dd$Deta2)
         good[good] <- dd$Deta2[good]>0 
-        w <- dd$Deta2*.5; w[!good] <- 0
-        thresh <- max(w[good])*.Machine$double.eps^.5
-        w[w < thresh] <- thresh
-        good <- is.finite(dd$Deta)
-        z <- (eta-offset)[good] - .5 * dd$Deta[good] / w[good]
+        #w <- dd$Deta2*.5; 
+        w[!good] <- 0
+        wz <- w*(eta-offset) - .5*dd$Deta
+        z <- (eta-offset) - dd$Deta.Deta2
+        good <- is.finite(z)&is.finite(w) 
+        if (sum(!good)) {
+          use.wy <- TRUE
+          good <- is.finite(w)&is.finite(wz)
+          z[!is.finite(z)] <- 0 ## avoid NaN in .C call - unused anyway
+        } else use.wy <- family$use.wz
+        #thresh <- max(w[good])*.Machine$double.eps^.5
+        #w[w < thresh] <- thresh
+        #good <- is.finite(dd$Deta)
+        #z <- (eta-offset)[good] - .5 * dd$Deta[good] / w[good]
        
         oo <- .C(C_pls_fit1, ##C_pls_fit1,
-                  y=as.double(z),X=as.double(x[good,]),w=as.double(w),
+                  y=as.double(z[good]),X=as.double(x[good,]),w=as.double(w[good]),wy = as.double(wz[good]),
                      E=as.double(Sr),Es=as.double(Eb),n=as.integer(sum(good)),
                      q=as.integer(ncol(x)),rE=as.integer(rows.E),eta=as.double(z),
                      penalty=as.double(1),rank.tol=as.double(rank.tol),
-                     nt=as.integer(control$nthreads))
+                     nt=as.integer(control$nthreads),use.wy=as.integer(use.wy))
       }
       start <- oo$y[1:ncol(x)] ## current coefficient estimates
       penalty <- oo$penalty ## size of penalty
@@ -360,7 +395,7 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
           conv <- FALSE
           warning("Non-finite coefficients at iteration ", 
                   iter)
-          break
+          return(list(REML=NA)) ## return immediately signalling failure
       }        
      
       mu <- linkinv(eta <- eta + offset)
@@ -376,8 +411,8 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
             coefold <- null.coef
             etaold <- null.eta
          }
-         warning("Step size truncated due to divergence", 
-                     call. = FALSE)
+         #warning("Step size truncated due to divergence", 
+         #            call. = FALSE)
          ii <- 1
          while (!is.finite(dev)) {
                if (ii > control$maxit) 
@@ -397,8 +432,8 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
 
       ## now step halve if mu or eta are out of bounds... 
       if (!(valideta(eta) && validmu(mu))) {
-         warning("Step size truncated: out of bounds", 
-                  call. = FALSE)
+         #warning("Step size truncated: out of bounds", 
+         #         call. = FALSE)
          ii <- 1
          while (!(valideta(eta) && validmu(mu))) {
                   if (ii > control$maxit) 
@@ -448,7 +483,7 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
        ## Need to check coefs converged adequately, to ensure implicit differentiation
        ## ok. Testing coefs unchanged is problematic under rank deficiency (not guaranteed to
        ## drop same parameter every iteration!)       
-       grad <- 2 * t(x[good,])%*%(w*((x%*%start)[good]-z))+ 2*St%*%start 
+       grad <- 2 * t(x[good,])%*%((w[good]*(x%*%start)[good]-wz[good]))+ 2*St%*%start 
        if (max(abs(grad)) > control$epsilon*max(abs(start+coefold))/2) {
       ## if (max(abs(start-coefold))>control$epsilon*max(abs(start+coefold))/2) {
          old.pdev <- pdev  ## not converged quite enough
@@ -480,26 +515,37 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
 
    dd <- dDeta(y,mu,weights,theta,family,deriv)
    w <- dd$Deta2 * .5
+   z <- (eta-offset) - dd$Deta.Deta2 ## - .5 * dd$Deta[good] / w
+   wf <- dd$EDeta2 * .5 ## Fisher type weights 
+   wz <- w*(eta-offset) - 0.5*dd$Deta ## Wz finite when w==0
+  
+   gdi.type <- if (any(abs(w)<.Machine$double.xmin*1e20)||any(!is.finite(z))) 1 else 0   
+   good <- is.finite(wz)&is.finite(w)   
+      
 
    ## exclude points for which gradient and second deriv are effectively zero and 
    ## points with non finite second deriv or deriv ratio... 
-   min.Deta <- mean(abs(dd$Deta[is.finite(dd$Deta)]))*.Machine$double.eps*.001
-   min.Deta2 <- mean(abs(dd$Deta2[is.finite(dd$Deta2)]))*.Machine$double.eps*.001
-   good <- is.finite(dd$Deta.Deta2)&is.finite(w)&!(abs(dd$Deta2) < min.Deta2 & abs(dd$Deta) < min.Deta) 
-   if (control$trace&sum(!good)>0) cat("\n",sum(!good)," not good\n")
-   w <- w[good] 
+   #min.Deta <- mean(abs(dd$Deta[is.finite(dd$Deta)]))*.Machine$double.eps*.001
+   #min.Deta2 <- mean(abs(dd$Deta2[is.finite(dd$Deta2)]))*.Machine$double.eps*.001
+   #good <- is.finite(dd$Deta.Deta2)&is.finite(w)&!(abs(dd$Deta2) < min.Deta2 & abs(dd$Deta) < min.Deta) 
+   #if (control$trace&sum(!good)>0) cat("\n",sum(!good)," not good\n")
+   #w <- w[good] 
 
-   z <- (eta-offset)[good] - dd$Deta.Deta2[good] ## - .5 * dd$Deta[good] / w
-   wf <- dd$EDeta2[good] * .5 ## Fisher type weights 
+   #z <- (eta-offset)[good] - dd$Deta.Deta2[good] ## - .5 * dd$Deta[good] / w
+   #wf <- dd$EDeta2[good] * .5 ## Fisher type weights 
+   #wz <- w*(eta-offset)[good] - 0.5*dd$Deta[good]
 
-   residuals <- rep.int(NA, nobs)
-   residuals[good] <- z - (eta - offset)[good]
+   #residuals <- rep.int(NA, nobs)
+   residuals <- z - (eta - offset)
+   residuals[!is.finite(residuals)] <- NA 
+   z[!is.finite(z)] <- 0 ## avoid passing NA etc to C code  
 
    ntot <- length(theta) + length(sp)
    ## if (deriv>1) n2d <- ntot*(1+ntot)/2 else n2d <- 0 
    rSncol <- unlist(lapply(UrS,ncol))
    ## Now drop any elements of dd that have been dropped in fitting...
-   if (sum(!good)>0) { ## drop !good from fields of dd
+   if (sum(!good)>0) { ## drop !good from fields of dd, weights and pseudodata
+     z <- z[good]; w <- w[good]; wz <- wz[good]; wf <- wf[good]
      dd$Deta <- dd$Deta[good];dd$Deta2 <- dd$Deta2[good] 
      dd$EDeta2 <- dd$EDeta2[good]
      if (deriv>0) dd$Deta3 <- dd$Deta3[good]
@@ -524,18 +570,19 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
        } 
      }
    }
-   ## can't have zero weights in gdi2 call
-   mwb <- max(abs(w))*.Machine$double.eps
-   mwa <- min(abs(w[w!=0]))*.0001; if (mwa==0) mwa <- mwb
-   w[w==0] <- min(mwa,mwb);
+   ## can't have zero weights in gdi2 call - superceded by type=1 handling of w==0
+   #mwb <- max(abs(w))*.Machine$double.eps
+   #mwa <- min(abs(w[w!=0]))*.0001; if (mwa==0) mwa <- mwb
+   #w[w==0] <- min(mwa,mwb);
    oo <- .C(C_gdi2,
             X=as.double(x[good,]),E=as.double(Sr),Es=as.double(Eb),rS=as.double(unlist(rS)),
             U1 = as.double(U1),sp=as.double(exp(sp)),theta=as.double(theta),
-            z=as.double(z),w=as.double(w),wf=as.double(wf),Dth=as.double(dd$Dth),Det=as.double(dd$Deta),
+            z=as.double(z),w=as.double(w),wz=as.double(wz),wf=as.double(wf),Dth=as.double(dd$Dth),
+            Det=as.double(dd$Deta),
             Det2=as.double(dd$Deta2),Dth2=as.double(dd$Dth2),Det.th=as.double(dd$Detath),
             Det2.th=as.double(dd$Deta2th),Det3=as.double(dd$Deta3),Det.th2 = as.double(dd$Detath2),
             Det4 = as.double(dd$Deta4),Det3.th=as.double(dd$Deta3th), Deta2.th2=as.double(dd$Deta2th2),
-            beta=as.double(coef),b1=as.double(rep(0,ntot*ncol(x))),
+            beta=as.double(coef),b1=as.double(rep(0,ntot*ncol(x))),w1=rep(0,ntot*length(z)),
             D1=as.double(rep(0,ntot)),D2=as.double(rep(0,ntot^2)),
             P=as.double(0),P1=as.double(rep(0,ntot)),P2 = as.double(rep(0,ntot^2)),
             ldet=as.double(1-2*(scoreType=="ML")),ldet1 = as.double(rep(0,ntot)), 
@@ -545,12 +592,34 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
 	    n=as.integer(sum(good)),q=as.integer(ncol(x)),M=as.integer(nSp),
             n.theta=as.integer(length(theta)), Mp=as.integer(Mp),Enrow=as.integer(rows.E),
             rSncol=as.integer(rSncol),deriv=as.integer(deriv),
-	    fixed.penalty = as.integer(rp$fixed.penalty),nt=as.integer(control$nthreads))
-
+	    fixed.penalty = as.integer(rp$fixed.penalty),nt=as.integer(control$nthreads),
+            type=as.integer(gdi.type),dVkk=as.double(rep(0,nSp^2)))
+## test code used to ensure type 0 and type 1 produce identical results, when both should work. 
+#   oot <- .C(C_gdi2,
+#            X=as.double(x[good,]),E=as.double(Sr),Es=as.double(Eb),rS=as.double(unlist(rS)),
+#            U1 = as.double(U1),sp=as.double(exp(sp)),theta=as.double(theta),
+#            z=as.double(z),w=as.double(w),wz=as.double(wz),wf=as.double(wf),Dth=as.double(dd$Dth),
+#            Det=as.double(dd$Deta),
+#            Det2=as.double(dd$Deta2),Dth2=as.double(dd$Dth2),Det.th=as.double(dd$Detath),
+#            Det2.th=as.double(dd$Deta2th),Det3=as.double(dd$Deta3),Det.th2 = as.double(dd$Detath2),
+#            Det4 = as.double(dd$Deta4),Det3.th=as.double(dd$Deta3th), Deta2.th2=as.double(dd$Deta2th2),
+#            beta=as.double(coef),b1=as.double(rep(0,ntot*ncol(x))),w1=rep(0,ntot*length(z)),
+#            D1=as.double(rep(0,ntot)),D2=as.double(rep(0,ntot^2)),
+#            P=as.double(0),P1=as.double(rep(0,ntot)),P2 = as.double(rep(0,ntot^2)),
+#            ldet=as.double(1-2*(scoreType=="ML")),ldet1 = as.double(rep(0,ntot)), 
+#            ldet2 = as.double(rep(0,ntot^2)),
+#            rV=as.double(rep(0,ncol(x)^2)),
+#            rank.tol=as.double(.Machine$double.eps^.75),rank.est=as.integer(0),
+#	    n=as.integer(sum(good)),q=as.integer(ncol(x)),M=as.integer(nSp),
+#            n.theta=as.integer(length(theta)), Mp=as.integer(Mp),Enrow=as.integer(rows.E),
+#            rSncol=as.integer(rSncol),deriv=as.integer(deriv),
+#	    fixed.penalty = as.integer(rp$fixed.penalty),nt=as.integer(control$nthreads),
+#            type=as.integer(1))
    rV <- matrix(oo$rV,ncol(x),ncol(x)) ## rV%*%t(rV)*scale gives covariance matrix 
    rV <- T %*% rV   
    ## derivatives of coefs w.r.t. sps etc...
    db.drho <- if (deriv) T %*% matrix(oo$b1,ncol(x),ntot) else NULL 
+   dw.drho <- if (deriv) matrix(oo$w1,length(z),ntot) else NULL
    Kmat <- matrix(0,nrow(x),ncol(x)) 
    Kmat[good,] <- oo$X                    ## rV%*%t(K)%*%(sqrt(wf)*X) = F; diag(F) is edf array 
 
@@ -600,13 +669,18 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
 
    names(coef) <- xnames
    names(residuals) <- ynames
-   wtdmu <- sum(weights * y)/sum(weights)
+   wtdmu <- sum(weights * mu)/sum(weights) ## changed from y
    nulldev <- sum(dev.resids(y, rep(wtdmu,length(y)), weights))
    n.ok <- nobs - sum(weights == 0)
    nulldf <- n.ok
    ww <- wt <- rep.int(0, nobs)
    wt[good] <- wf 
    ww[good] <- w
+   if (deriv && nrow(dw.drho)!=nrow(x)) {
+      w1 <- dw.drho
+      dw.drho <- matrix(0,nrow(x),ncol(w1))
+      dw.drho[good,] <- w1
+   }
    aic.model <- family$aic(y, mu, theta, weights, dev) # note: incomplete 2*edf needs to be added
  
 
@@ -619,11 +693,12 @@ gam.fit4 <- function(x, y, sp, Eb,UrS=list(),
         df.null = nulldf, y = y, converged = conv,
         boundary = boundary,
         REML=REML,REML1=REML1,REML2=REML2,
-        rV=rV,db.drho=db.drho,
+        rV=rV,db.drho=db.drho,dw.drho=dw.drho,
         scale.est=scale,reml.scale=scale,
         aic=aic.model,
         rank=oo$rank.est,
-        K=Kmat,control=control
+        K=Kmat,control=control,
+        dVkk = matrix(oo$dVkk,nSp,nSp)
         #,D1=oo$D1,D2=D2,
         #ldet=oo$ldet,ldet1=oo$ldet1,ldet2=ldet2,
         #bSb=oo$P,bSb1=oo$P1,bSb2=bSb2,
@@ -656,10 +731,7 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
   penalized <- if (length(Sl)>0) TRUE else FALSE
 
   nSp <- length(lsp)
-  ##sp <- exp(lsp) 
-  ## rank.tol <- .Machine$double.eps*100 ## tolerance to use for rank deficiency
   q <- ncol(x)
-  ##n <- 
   nobs <- length(y)
   
   if (penalized) {
@@ -699,8 +771,9 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
  
 
   ## get log likelihood, grad and Hessian (w.r.t. coefs - not s.p.s) ...
-  ll <- family$ll(y,x,coef,weights,family,deriv=1) 
-  ll0 <- ll$l - t(coef)%*%St%*%coef/2
+  llf <- family$ll
+  ll <- llf(y,x,coef,weights,family,offset=offset,deriv=1) 
+  ll0 <- ll$l - (t(coef)%*%St%*%coef)/2
   rank.checked <- FALSE ## not yet checked the intrinsic rank of problem 
   rank <- q;drop <- NULL
   eigen.fix <- FALSE
@@ -714,7 +787,7 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
       fdg <- ll$lb*0; fdh <- ll$lbb*0
       for (k in 1:length(coef)) {
         coef1 <- coef;coef1[k] <- coef[k] + eps
-        ll.fd <- family$ll(y,x,coef1,weights,family,deriv=1)
+        ll.fd <- llf(y,x,coef1,weights,family,offset=offset,deriv=1)
         fdg[k] <- (ll.fd$l-ll$l)/eps
         fdh[,k] <- (ll.fd$lb-ll$lb)/eps
       }
@@ -728,8 +801,6 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
       if (eigen.fix) {
         eh <- eigen(Hp,symmetric=TRUE);
         ev <- abs(eh$values)
-        #thresh <- min(ev[ev>0])
-        #ev[ev<thresh] <- thresh
         Hp <- eh$vectors%*%(ev*t(eh$vectors))
       } else {
         Ib <- diag(rank)*abs(min(D))
@@ -771,18 +842,38 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
     }
     ## try the Newton step...
     coef1 <- coef + step 
-    ll <- family$ll(y,x,coef1,weights,family,deriv=1) 
+    ll <- llf(y,x,coef1,weights,family,offset=offset,deriv=1) 
     ll1 <- ll$l - (t(coef1)%*%St%*%coef1)/2
-    khalf <- 0
-    while (ll1 < ll0 && khalf < 50) { ## step halve until it succeeds...
-      step <- step/2;coef1 <- coef + step
-      ll <- family$ll(y,x,coef1,weights,family,deriv=0)
+    khalf <- 0;fac <- 2
+    while (ll1 < ll0 && khalf < 25) { ## step halve until it succeeds...
+      step <- step/fac;coef1 <- coef + step
+      ll <- llf(y,x,coef1,weights,family,offset=offset,deriv=0)
       ll1 <- ll$l - (t(coef1)%*%St%*%coef1)/2
       if (ll1>=ll0) {
-        ll <- family$ll(y,x,coef1,weights,family,deriv=1)
+        ll <- llf(y,x,coef1,weights,family,offset=offset,deriv=1)
+      } else { ## abort if step has made no difference
+        if (max(abs(coef1-coef))==0) khalf <- 100
       }
       khalf <- khalf + 1
+      if (khalf>5) fac <- 5
     } ## end step halve
+ 
+    if (ll1 < ll0) { ## switch to steepest descent... 
+      step <- -.5*drop(grad)*mean(abs(coef))/mean(abs(grad))
+      khalf <- 0
+    }
+
+    while (ll1 < ll0 && khalf < 25) { ## step cut until it succeeds...
+      step <- step/10;coef1 <- coef + step
+      ll <- llf(y,x,coef1,weights,family,offset=offset,deriv=0)
+      ll1 <- ll$l - (t(coef1)%*%St%*%coef1)/2
+      if (ll1>=ll0) {
+        ll <- llf(y,x,coef1,weights,family,offset=offset,deriv=1)
+      } else { ## abort if step has made no difference
+        if (max(abs(coef1-coef))==0) khalf <- 100
+      }
+      khalf <- khalf + 1
+    }
 
     if (ll1 >= ll0||iter==control$maxit) { ## step ok. Accept and test
       coef <- coef + step
@@ -796,7 +887,7 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
             perturbed <- perturbed + 1
             coef <- coef*(1+(runif(length(coef))*.02-.01)*perturbed) + 
                     (runif(length(coef)) - 0.5 ) * mean(abs(coef))*1e-5*perturbed 
-            ll <- family$ll(y,x,coef,weights,family,deriv=1) 
+            ll <- llf(y,x,coef,weights,family,offset=offset,deriv=1) 
             ll0 <- ll$l - (t(coef)%*%St%*%coef)/2
           } else {        
             rank.checked <- TRUE
@@ -816,23 +907,23 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
               bdrop <- 1:q %in% drop ## TRUE FALSE version
               ## now drop the parameters and recompute ll0...
               lpi <- attr(x,"lpi")
+              xat <- attributes(x)
+              xat$dim <- xat$dimnames <- NULL
               coef <- coef[-drop]
               St <- St[-drop,-drop]
               x <- x[,-drop] ## dropping columns from model matrix
               if (!is.null(lpi)) { ## need to adjust column indexes as well
                 ii <- (1:q)[!bdrop];ij <- rep(NA,q)
                 ij[ii] <- 1:length(ii) ## col i of old model matrix is col ij[i] of new 
-                #k <- 0
+               
                 for (i in 1:length(lpi)) {
-                  #kk <- sum(lpi[[i]]%in%drop==FALSE) ## how many left undropped?
-                  #lpi[[i]] <- 1:kk + k ## new index - note strong assumptions on structure here
-                  #k <- k + kk
                   lpi[[i]] <- ij[lpi[[i]][!(lpi[[i]]%in%drop)]] # drop and shuffle up
                 }
               } ## lpi adjustment done
+              for (i in 1:length(xat)) attr(x,names(xat)[i]) <- xat[[i]]
               attr(x,"lpi") <- lpi
               attr(x,"drop") <- drop ## useful if family has precomputed something from x
-              ll <- family$ll(y,x,coef,weights,family,deriv=1) 
+              ll <- llf(y,x,coef,weights,family,offset=offset,deriv=1) 
               ll0 <- ll$l - (t(coef)%*%St%*%coef)/2
             } 
           }
@@ -861,7 +952,7 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
     fcoef <- rep(0,length(bdrop));fcoef[!bdrop] <- coef
   } else fcoef <- coef
 
-  d1l <- d2l <- d1bSb <- d2bSb <- d1b <- d2b <- d1ldetH <- d2ldetH <- d1b <- d2b <- NULL
+  dVkk <- d1l <- d2l <- d1bSb <- d2bSb <- d1b <- d2b <- d1ldetH <- d2ldetH <- d1b <- d2b <- NULL
 
   if (deriv>0) {  ## Implicit differentiation for derivs...
 
@@ -871,6 +962,9 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
     if (nSp) for (i in 1:m) d1b[,i] <- 
        -D*(backsolve(L,forwardsolve(t(L),(D*Sib[[i]][!bdrop])[piv]))[ipiv])
   
+    ## obtain the curvature check matrix...
+    dVkk <- crossprod(L[,ipiv]%*%(d1b/D))
+
     if (!is.null(drop)) { ## create full version of d1b with zeros for unidentifiable 
       fd1b <-  matrix(0,q,m)
       fd1b[!bdrop,] <- d1b
@@ -879,8 +973,8 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
     ## Now call the family again to get first derivative of Hessian w.r.t
     ## smoothing parameters, in list d1H...
 
-    ll <- family$ll(y,x,coef,weights,family,deriv=3,d1b=d1b)
-    d1l <- colSums(ll$lb*d1b)
+    ll <- llf(y,x,coef,weights,family,offset=offset,deriv=3,d1b=d1b)
+    # d1l <- colSums(ll$lb*d1b) # cancels
     
 
     if (deriv>1) { ## Implicit differentiation for the second derivatives is now possible...
@@ -896,17 +990,18 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
   
       ## Now call family for last time to get trHid2H the tr(H^{-1} d^2 H / drho_i drho_j)...
 
-      llr <- family$ll(y,x,coef,weights,family,deriv=4,d1b=d1b,d2b=d2b,
+      llr <- llf(y,x,coef,weights,family,offset=offset,deriv=4,d1b=d1b,d2b=d2b,
                        Hp=Hp,rank=rank,fh = L,D=D)
 
       ## Now compute Hessian of log lik w.r.t. log sps using chain rule
        
-      d2la <- colSums(ll$lb*d2b)
-      k <- 0
+      # d2la <- colSums(ll$lb*d2b) # cancels
+      # k <- 0
       d2l <- matrix(0,m,m)
       for (i in 1:m) for (j in i:m) {
-        k <- k + 1
-        d2l[j,i] <- d2l[i,j] <- d2la[k] + t(d1b[,i])%*%ll$lbb%*%d1b[,j] 
+        # k <- k + 1
+        d2l[j,i] <- d2l[i,j] <- # d2la[k] + # cancels
+	                        t(d1b[,i])%*%ll$lbb%*%d1b[,j] 
       }
     } ## if (deriv > 1)
   } ## if (deriv > 0)
@@ -929,11 +1024,14 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
       k <- k + 1
       d2ldetH[i,j] <- -sum(d1Hp[[i]]*t(d1Hp[[j]])) - llr$trHid2H[k] 
       if (i==j) { ## need to add term relating to smoothing penalty
-        A <- t(Sl.mult(rp$Sl,diag(q),i,full=FALSE))
-        bind <- rowSums(A)!=0
-        ind <- which(bind)
-        bind <- bind[!bdrop]
-        A <- A[!bdrop,!bdrop[ind]]
+        #A <- t(Sl.mult(rp$Sl,diag(q),i,full=FALSE))
+        #bind <- rowSums(abs(A))!=0 ## FIX: abs 3/3/16 
+        #ind <- which(bind)
+        #bind <- bind[!bdrop]
+        #A <- A[!bdrop,!bdrop[ind]]
+        A <- Sl.mult(rp$Sl,diag(q),i,full=TRUE)[!bdrop,!bdrop]
+        bind <- rowSums(abs(A))!=0 ## row/cols of non-zero block
+        A <- A[,bind] ## drop the zero columns  
         A <- D*(backsolve(L,forwardsolve(t(L),(D*A)[piv,]))[ipiv,])
         d2ldetH[i,j] <- d2ldetH[i,j] + sum(diag(A[bind,]))
       } else d2ldetH[j,i] <- d2ldetH[i,j]
@@ -943,23 +1041,24 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
   ## Compute derivs of b'Sb...
 
   if (deriv>0) {
-    Sb <- St%*%coef
+    # Sb <- St%*%coef
     Skb <- Sl.termMult(rp$Sl,fcoef,full=TRUE)
     d1bSb <- rep(0,m)
     for (i in 1:m) { 
       Skb[[i]] <- Skb[[i]][!bdrop]
-      d1bSb[i] <- 2*sum(d1b[,i]*Sb) + sum(coef*Skb[[i]])
+      d1bSb[i] <- # 2*sum(d1b[,i]*Sb) + # cancels
+                  sum(coef*Skb[[i]])
     }
   }
  
   if (deriv>1) {
     d2bSb <- matrix(0,m,m)
-    k <- 0
+    # k <- 0
     for (i in 1:m) {
       Sd1b <- St%*%d1b[,i] 
       for (j in i:m) {
-        k <- k + 1
-        d2bSb[j,i] <- d2bSb[i,j] <- 2*sum(d2b[,k]*Sb + 
+         k <- k + 1
+         d2bSb[j,i] <- d2bSb[i,j] <- 2*sum( # d2b[,k]*Sb + # cancels 
          d1b[,i]*Skb[[j]] + d1b[,j]*Skb[[i]] + d1b[,j]*Sd1b)
       }
       d2bSb[i,i] <-  d2bSb[i,i] + sum(coef*Skb[[i]]) 
@@ -967,30 +1066,34 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
   }
 
   ## get grad and Hessian of REML score...
-  REML <- -as.numeric(ll$l - t(coef)%*%St%*%coef/2 + rp$ldetS/2 - ldetHp/2 + Mp*log(2*pi)/2)
+  REML <- -as.numeric(ll$l - t(coef)%*%St%*%coef/2 + rp$ldetS/2  - ldetHp/2  + Mp*log(2*pi)/2)
  
-  REML1 <- if (deriv>0) -as.numeric(d1l - d1bSb/2 + rp$ldet1/2 - d1ldetH/2) else NULL 
+  REML1 <- if (deriv<1) NULL else -as.numeric( # d1l # cancels
+                                   - d1bSb/2 + rp$ldet1/2  - d1ldetH/2 ) 
+
   if (control$trace) {
     cat("\niter =",iter,"  ll =",ll$l,"  REML =",REML,"  bSb =",t(coef)%*%St%*%coef/2,"\n")
     cat("log|S| =",rp$ldetS,"  log|H+S| =",ldetHp,"  n.drop =",length(drop),"\n")
     if (!is.null(REML1)) cat("REML1 =",REML1,"\n")
   }
-  REML2 <- if (deriv>1) -(d2l - d2bSb/2 + rp$ldet2/2 - d2ldetH/2) else NULL 
+  REML2 <- if (deriv<2) NULL else -( d2l - d2bSb/2 + rp$ldet2/2  - d2ldetH/2 ) 
  ## bSb <- t(coef)%*%St%*%coef
   lpi <- attr(x,"lpi")
   if (is.null(lpi)) { 
-    linear.predictors <- as.numeric(x%*%coef)
+    linear.predictors <- if (is.null(offset)) as.numeric(x%*%coef) else as.numeric(x%*%coef+offset)
     fitted.values <- family$linkinv(linear.predictors) 
   } else {
     fitted.values <- linear.predictors <- matrix(0,nrow(x),length(lpi))
+    if (!is.null(offset)) offset[[length(lpi)+1]] <- 0
     for (j in 1:length(lpi)) {
       linear.predictors[,j] <- as.numeric(x[,lpi[[j]],drop=FALSE] %*% coef[lpi[[j]]])
+      if (!is.null(offset[[j]])) linear.predictors[,j] <-  linear.predictors[,j] + offset[[j]]
       fitted.values[,j] <- family$linfo[[j]]$linkinv( linear.predictors[,j]) 
     }
   }
   coef <- Sl.repara(rp$rp,fcoef,inverse=TRUE) ## undo re-parameterization of coef 
  
-  if (!is.null(drop)) { ## create full version of d1b with zeros for unidentifiable 
+  if (!is.null(drop)&&!is.null(d1b)) { ## create full version of d1b with zeros for unidentifiable 
     db.drho <- matrix(0,length(bdrop),ncol(d1b));db.drho[!bdrop,] <- d1b
   } else db.drho <- d1b
   ## and undo re-para...
@@ -1001,7 +1104,8 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
        scale.est=1, ### NOTE: needed by newton, but what is sensible here? 
        REML= REML,REML1= REML1,REML2=REML2,
        rank=rank,aic = -2*ll$l, ## 2*edf needs to be added
-       l= ll$l,l1 =d1l,l2 =d2l,
+       ##deviance = -2*ll$l,
+       l= ll$l,## l1 =d1l,l2 =d2l,
        lbb = ll$lbb, ## Hessian of log likelihood
        L=L, ## chol factor of pre-conditioned penalized hessian
        bdrop=bdrop, ## logical index of dropped parameters
@@ -1010,14 +1114,104 @@ gam.fit5 <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,deriv=2,family,
        rp = rp$rp,
        db.drho = db.drho, ## derivative of penalty coefs w.r.t. log sps.
        #bSb = bSb, bSb1 =  d1bSb,bSb2 =  d2bSb,
+       S1=rp$ldet1,
        #S=rp$ldetS,S1=rp$ldet1,S2=rp$ldet2,
        #Hp=ldetHp,Hp1=d1ldetH,Hp2=d2ldetH,
        #b2 = d2b)
-       H = ll$lbb,dH = ll$d1H)#,d2H=llr$d2H)
+       niter=iter,H = ll$lbb,dH = ll$d1H,dVkk=dVkk)#,d2H=llr$d2H)
+    ## debugging code to allow components of 2nd deriv of hessian w.r.t. sp.s 
+    ## to be passed to deriv.check.... 
+    #if (!is.null(ll$ghost1)&&!is.null(ll$ghost2)) { 
+    #  ret$ghost1 <- ll$ghost1; ret$ghost2 <- ret$ghost2
+    #} 
     ret
 } ## end of gam.fit5
 
-gam.fit5.post.proc <- function(object,Sl,L) {
+efsud <- function(x,y,lsp,Sl,weights=NULL,offset=NULL,family,
+                     control=gam.control(),Mp=-1,start=NULL) {
+## Extended Fellner-Schall method
+## tr(S^-S_j) is returned by ldetS as ldet1 - S1 from gam.fit5
+## b'S_jb is computed as d1bSb in gam.fit5
+## tr(V S_j) will need to be computed using Sl.termMult
+##   Sl returned by ldetS and Vb computed as in gam.fit5.postproc.
+  tol <- 1e-6
+  lsp <- lsp + 2.5
+  mult <- 1
+  fit <- gam.fit5(x=x,y=y,lsp=lsp,Sl=Sl,weights=weights,offset=offset,deriv=0,family=family,
+                     control=control,Mp=Mp,start=start)
+  score.hist <- rep(0,200)
+  for (iter in 1:200) {
+    start <- fit$coefficients
+    ## obtain Vb...
+    ipiv <- piv <- attr(fit$L,"pivot")
+    p <- length(piv)
+    ipiv[piv] <- 1:p
+    Vb <- crossprod(forwardsolve(t(fit$L),diag(fit$D,nrow=p)[piv,,drop=FALSE])[ipiv,,drop=FALSE])
+    if (sum(fit$bdrop)) { ## some coefficients were dropped...
+      q <- length(fit$bdrop)
+      ibd <- !fit$bdrop
+      Vtemp <- Vb; Vb <- matrix(0,q,q)
+      Vb[ibd,ibd] <- Vtemp
+    }
+    Vb <- Sl.repara(fit$rp,Vb,inverse=TRUE)
+    SVb <- Sl.termMult(Sl,Vb) ## this could be made more efficient
+    trVS <- rep(0,length(SVb))
+    for (i in 1:length(SVb)) {
+      ind <- attr(SVb[[i]],"ind")
+      trVS[i] <- sum(diag(SVb[[i]][,ind]))
+    }
+    Sb <- Sl.termMult(Sl,start,full=TRUE)
+    bSb <- rep(0,length(Sb))
+    for (i in 1:length(Sb)) {
+      bSb[i] <- sum(start*Sb[[i]])
+    }
+    a <- pmax(0,fit$S1*exp(-lsp) - trVS)
+    r <- a/pmax(0,bSb)
+    r[a==0&bSb==0] <- 1
+    r[!is.finite(r)] <- 1e6
+    lsp1 <- pmin(lsp + log(r)*mult,12)
+    old.reml <- fit$REML
+    fit <- gam.fit5(x=x,y=y,lsp=lsp1,Sl=Sl,weights=weights,offset=offset,deriv=0,
+                    family=family,control=control,Mp=Mp,start=start)
+    ## some step length control...
+   
+    if (fit$REML<=old.reml) { ## improvement
+      if (max(abs(log(r))<.05)) { ## consider step extension
+        lsp2 <- pmin(lsp + log(r)*mult*2,12) ## try extending step...
+        fit2 <- gam.fit5(x=x,y=y,lsp=lsp2,Sl=Sl,weights=weights,offset=offset,deriv=0,family=family,
+                     control=control,Mp=Mp,start=start)
+     
+        if (fit2$REML < fit$REML) { ## improvement - accept extension
+          fit <- fit2;lsp <- lsp2
+	  mult <- mult * 2
+        } else { ## accept old step
+          lsp <- lsp1
+        }
+      } else lsp <- lsp1
+    } else { ## no improvement 
+      while (fit$REML > old.reml&&mult>1) { ## don't contract below 1 as update doesn't have to improve REML 
+          mult <- mult/2 ## contract step
+          lsp1 <- pmin(lsp + log(r)*mult,12)
+	  fit <- gam.fit5(x=x,y=y,lsp=lsp1,Sl=Sl,weights=weights,offset=offset,deriv=0,family=family,
+                        control=control,Mp=Mp,start=start)
+      }
+      lsp <- lsp1
+      if (mult<1) mult <- 1
+    }
+    score.hist[iter] <- fit$REML
+    if (iter==1) old.ll <- fit$l else {
+      if (abs(old.ll-fit$l)<tol*abs(fit$l)) break
+      old.ll <- fit$l
+    }
+  }
+  fit$sp <- exp(lsp)
+  fit$niter <- iter
+  fit$outer.info <- list(iter = iter,score.hist=score.hist[1:iter])
+  fit$outer.info$conv <- if (iter==200) "iteration limit reached" else "full convergence"
+  fit
+} ## efsud
+
+gam.fit5.post.proc <- function(object,Sl,L,lsp0,S,off) {
 ## object is object returned by gam.fit5, Sl is penalty object, L maps working sp
 ## vector to full sp vector 
 ## Computes:
@@ -1089,13 +1283,27 @@ gam.fit5.post.proc <- function(object,Sl,L) {
   }  
 
   ## compute the smoothing parameter uncertainty correction...
-  if (!is.null(object$outer.info$hess)) {
-    ev <- eigen(object$outer.info$hess,symmetric=TRUE)
-    ind <- ev$values <= 0
-    ev$values[ind] <- 0;ev$values[!ind] <- 1/sqrt(ev$values[!ind])
+  if (!is.null(object$outer.info$hess)&&!is.null(object$db.drho)) { 
     if (!is.null(L)) object$db.drho <- object$db.drho%*%L ## transform to derivs w.r.t. working
-    Vc <- crossprod((ev$values*t(ev$vectors))%*%t(object$db.drho))
-    Vc <- Vb + Vc  ## Bayesian cov matrix with sp uncertainty
+    ev <- eigen(object$outer.info$hess,symmetric=TRUE)
+    d <- ev$values;ind <- d <= 0
+    d[ind] <- 0;d[!ind] <- 1/sqrt(d[!ind])
+    Vc <- crossprod((d*t(ev$vectors))%*%t(object$db.drho))
+    #dpv <- rep(0,ncol(object$outer.info$hess));M <- length(off)
+    #dpv[1:M] <- 1/100 ## prior precision (1/var) on log smoothing parameters
+    #Vr <- chol2inv(chol(object$outer.info$hess + diag(dpv,ncol=length(dpv))))[1:M,1:M]
+    #Vc <- object$db.drho%*%Vr%*%t(object$db.drho)
+    
+    #dpv[1:M] <- 1/10 ## prior precision (1/var) on log smoothing parameters
+    #Vr <- chol2inv(chol(object$outer.info$hess + diag(dpv,ncol=length(dpv))))[1:M,1:M]
+    #M <- length(off)
+    d <- ev$values; d[ind] <- 0;
+    d <- d + 1/50 #d[1:M] <- d[1:M] + 1/50 
+    d <- 1/sqrt(d)
+    Vr <- crossprod(d*t(ev$vectors))
+    #Vc2 <- Vb.corr(R,L,S,off,dw=NULL,w=NULL,log(object$sp),Vr)
+
+    Vc <- Vb + Vc #+ Vc2  ## Bayesian cov matrix with sp uncertainty
     ## reverse the various re-parameterizations...
   } else Vc <- Vb
   Vc <- Sl.repara(object$rp,Vc,inverse=TRUE) 
@@ -1112,8 +1320,14 @@ gam.fit5.post.proc <- function(object,Sl,L) {
   ## model. This is larger than edf2 should be, because of bias correction variability,
   ##  but is bounded in a way that is not *guaranteed* for edf2. Note that 
   ## justification only applies to sum(edf1/2) not elementwise   
+  if (!is.null(object$outer.info$hess)&&!is.null(object$db.drho)) { 
+    ## second correction term is easier computed in original parameterization...
+    Vc2 <- Vb.corr(R,L,lsp0,S,off,dw=NULL,w=NULL,log(object$sp),Vr)
+    Vc <- Vc + Vc2
+  }
   edf1 <- 2*edf - rowSums(t(F)*F)
-  edf2 <- diag(Vc%*%crossprod(R)) 
+  #edf2 <- diag(Vc%*%crossprod(R)) 
+  edf2 <- rowSums(Vc*crossprod(R))
   if (sum(edf2)>sum(edf1)) edf2 <- edf1 
   ## note hat not possible here...
   list(Vc=Vc,Vb=Vb,Ve=Ve,edf=edf,edf1=edf1,edf2=edf2,F=F,R=R)
